mod utils;

use std::convert::TryInto;

use wasm_bindgen::prelude::*;
use wordlelike::*;
use wordlelike::evaluator::Evaluator;

// When the `wee_alloc` feature is enabled, use `wee_alloc` as the global
// allocator.
#[cfg(feature = "wee_alloc")]
#[global_allocator]
static ALLOC: wee_alloc::WeeAlloc = wee_alloc::WeeAlloc::INIT;

#[macro_export]
macro_rules! matches_to_wordlelike {
    ( $x:expr ) => {
        {
            $x.iter()
                .map(|n| match n {
                    0 => wordlelike::guess::LetterMatch::NOWHERE,
                    1 => wordlelike::guess::LetterMatch::ELSEWHERE,
                    2 => wordlelike::guess::LetterMatch::HERE,
                    _ => wordlelike::guess::LetterMatch::NOWHERE,
                })
                .collect::<Vec<wordlelike::guess::LetterMatch>>()
                .try_into()
                .unwrap()
        }
    }
}

#[macro_export]
macro_rules! matches_from_wordlelike {
    ( $x:expr ) => {
        {
            $x.iter()
                .map(|n| match n {
                    wordlelike::guess::LetterMatch::NOWHERE => 0,
                    wordlelike::guess::LetterMatch::ELSEWHERE => 1,
                    wordlelike::guess::LetterMatch::HERE => 2,
                })
                .collect::<Vec<u8>>()
                .try_into()
                .unwrap()
        }
    }
}

#[wasm_bindgen]
#[derive(serde::Serialize, serde::Deserialize)]
pub struct GuessResult {
    word: String,
    matches: Vec<u8>,
}

#[wasm_bindgen]
impl GuessResult {
    pub fn word(&self) -> JsValue {
        serde_wasm_bindgen::to_value(&self.word).unwrap()
    }

    pub fn matches(&self) -> JsValue {
        serde_wasm_bindgen::to_value(&self.matches).unwrap()
    }
}

#[wasm_bindgen]
pub struct Game {
    dictionary: Vec<guess::Word<5>>,
    initial_history: Vec<GuessResult>,
    current_attempt: Vec<GuessResult>,
}

impl Game {
    fn build_adversary(&self) -> evaluator::AdversarialEvaluator<5> {
        let mut adversary = evaluator::AdversarialEvaluator::new(&self.dictionary);
        for result in &self.initial_history {
            adversary.push(guess::GuessResult {
                word: word!(result.word),
                matches: matches_to_wordlelike!(result.matches),
            });
        }
        for result in &self.current_attempt {
            adversary.push(guess::GuessResult {
                word: word!(result.word),
                matches: matches_to_wordlelike!(result.matches),
            });
        }
        adversary
    }
}

#[wasm_bindgen]
impl Game {
    pub fn new(dictionary: JsValue, initial_history: JsValue) -> Self {
        utils::set_panic_hook();
        let dict: Vec<String> = serde_wasm_bindgen::from_value(dictionary).unwrap();
        Self {
            dictionary: dict.iter().map(|word| word!(word)).collect(),
            initial_history: serde_wasm_bindgen::from_value(initial_history).unwrap(),
            current_attempt: Vec::new(),
        }
    }
    
    pub fn current_attempt(&self) -> JsValue {
        serde_wasm_bindgen::to_value(&self.current_attempt).unwrap()
    }

    pub fn guess(&mut self, word: String) -> JsValue {
        let word_: guess::Word<5> = word!(word);
        let adversary = self.build_adversary();
        let matches: Vec<u8> = matches_from_wordlelike!(adversary.evaluate(word_));
        self.current_attempt.push(GuessResult { word, matches: matches.to_vec() });
        
        serde_wasm_bindgen::to_value(&matches).unwrap()
    }

    pub fn possible_word(&self) -> String {
        let adversary = self.build_adversary();
        adversary.possible_word()
            .unwrap()
            .iter()
            .collect()
    }

    pub fn explore_game_tree(&self) -> JsValue {
        let mut adversary = evaluator::AdversarialEvaluator::new(&self.dictionary);
        for result in &self.initial_history {
            adversary.push(guess::GuessResult {
                word: word!(result.word),
                matches: matches_to_wordlelike!(result.matches),
            });
        }
        let first_word = &self.current_attempt.first().unwrap().word;
        let secrets = adversary.best_secrets(word![first_word]);
        let mut game_tree = Vec::new();
        for secret in secrets {
            game_tree.push(vec![
                GuessResult {
                    word: first_word.clone(),
                    matches: matches_from_wordlelike![guess::guess(word![first_word], &secret)],
                },
                GuessResult {
                    word: secret.iter().collect(),
                    matches: vec![2,2,2,2,2],
                }
            ]);
        }
        serde_wasm_bindgen::to_value(&game_tree).unwrap()
    }

    pub fn reset(&mut self) {
        self.current_attempt = Vec::new();
    }
}